@@ -15,6 +15,12 @@ def pytest_addoption(parser):
         default=100,
         help="asyncio: maximum number of tasks to run concurrently (int)",
     )
+    parser.addoption(
+        "--asyncio-task-timeout",
+        action="store",
+        default=120,
+        help="asyncio: number of seconds before a test will be cancelled (int)",
+    )
 
 
 def pytest_configure(config):
@@ -150,7 +156,7 @@ async def fill_fixture_fixtures(_fixtureinfo, fixture, item):
         )
 
         # Cache the module call
-        if fixture.scope in ["module"]:
+        if fixture.scope in ["module", "session"]:
             if not isinstance(fixture.func, CachedAsyncGen):
                 fixture.func = CachedAsyncGen(fixture.func)
 
@@ -166,7 +172,7 @@ async def fill_fixture_fixtures(_fixtureinfo, fixture, item):
         )
 
         # Cache the module call
-        if fixture.scope in ["module"]:
+        if fixture.scope in ["module", "session"]:
             if not isinstance(fixture.func, CachedFunction):
                 fixture.func = CachedFunction(fixture.func)
 
@@ -176,8 +182,11 @@ async def fill_fixture_fixtures(_fixtureinfo, fixture, item):
     # Regular fixture
     # FIXME: we should use more of pytest's fixture system
     elif inspect.isgeneratorfunction(fixture.func):
-        gen = fixture.func()
-        return gen.__next__(), [gen]
+        fixture_values, teardowns = await _fill_fixture_fixtures(
+            _fixtureinfo, fixture, item
+        )
+        gen = fixture.func(*fixture_values)
+        return gen.__next__(), teardowns + [gen]
 
     # It's a parameterization
     # FIXME: we should use more of pytest's fixture system
@@ -189,8 +198,11 @@ async def fill_fixture_fixtures(_fixtureinfo, fixture, item):
 
     # FIXME: we should use more of pytest's fixture system
     elif inspect.isfunction(fixture.func):
-        val = fixture.func()
-        return val, []
+        fixture_values, teardowns = await _fill_fixture_fixtures(
+            _fixtureinfo, fixture, item
+        )
+        val = fixture.func(*fixture_values)
+        return val, teardowns
 
     else:
         raise Exception(
@@ -216,7 +228,7 @@ async def fill_fixtures(item):
     for arg_name in function_args(item.function):
         fixture = _get_fixture(item._fixtureinfo, arg_name)
 
-        if fixture.scope not in ["function", "module"]:
+        if fixture.scope not in ["function", "module", "session"]:
             raise Exception(f"{fixture.scope} scope not supported")
 
         value, teardowns2 = await fill_fixture_fixtures(
@@ -359,12 +371,22 @@ def pytest_runtestloop(session):
         sidelined_tasks = tasks[max_tasks:]
         tasks = tasks[:max_tasks]
 
+        task_timeout = int(session.config.getoption("--asyncio-task-timeout"))
+
         completed = []
         while tasks:
             done, pending = await asyncio.wait(
-                tasks, return_when=asyncio.FIRST_COMPLETED
+                tasks, return_when=asyncio.FIRST_COMPLETED, timeout=5
             )
-            tasks = list(pending)
+
+            # Cancel tasks that have taken too long
+            tasks = []
+            for task in pending:
+                now = time.time()
+                item = item_by_coro[task._coro]
+                if hasattr(item, "start") and task_timeout < now - item.start:
+                    task.cancel()
+                tasks.append(task)
 
             for result in done:
                 item = item_by_coro[result._coro]
